# coding: utf-8

# docs
import asyncio
from typing import Callable, Dict, List, Optional, Union

from tqdm import tqdm

from supervisely._utils import batched
from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudAnnotation
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudAnnotationAPI(EntityAnnotationAPI):
    """
    :class:`PointcloudAnnotation<supervisely.pointcloud_annotation.pointcloud_annotation.PointcloudAnnotation>` for a single point cloud. :class:`PointcloudAnnotationAPI<PointcloudAnnotationAPI>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        pointcloud_id = 19618685
        ann_info = api.pointcloud.annotation.download(src_pointcloud_id)
    """

    _method_download_bulk = "point-clouds.annotations.bulk.info"
    _entity_ids_str = ApiField.POINTCLOUD_IDS

    def download(self, pointcloud_id: int) -> List[Dict]:
        """
        Download information about PointcloudAnnotation by point cloud ID from API.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :return: Information about PointcloudAnnotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pointcloud_id = 19618685
            ann_info = api.pointcloud.annotation.download(src_pointcloud_id)
            print(ann_info)

            # Output: {
            #     'datasetId': 62664,
            #     'description': '',
            #     'frames': [{'figures': [{'classId': None,
            #                             'createdAt': '2023-04-05T08:55:52.526Z',
            #                             'description': '',
            #                             'geometry': {'dimensions': {'x': 1.6056261,
            #                                                         'y': 3.8312221,
            #                                                         'z': 1.8019634},
            #                                         'position': {'x': 10.181290418102629,
            #                                                     'y': 9.033275311626847,
            #                                                     'z': -0.9065660704238034},
            #                                         'rotation': {'x': 0,
            #                                                     'y': 0,
            #                                                     'z': 1.5985649758590998}},
            #                             'geometryType': 'cuboid_3d',
            #                             'id': 87830573,
            #                             'labelerLogin': 'almaz',
            #                             'objectId': 5565738,
            #                             'updatedAt': '2023-04-05T08:55:52.526Z'},
            #                             {'classId': None,
            #                             'createdAt': '2023-04-05T08:55:52.526Z',
            #                             'description': '',
            #                             'geometry': {'dimensions': {'x': 2.3652234,
            #                                                         'y': 23.291742,
            #                                                         'z': 3.326648},
            #                                         'position': {'x': 77.40255111910977,
            #                                                     'y': -9.582723835261527,
            #                                                     'z': 1.0131292020311293},
            #                                         'rotation': {'x': 0,
            #                                                     'y': 0,
            #                                                     'z': -1.5823898471886868}},
            #                             'geometryType': 'cuboid_3d',
            #                             'id': 87830574,
            #                             'labelerLogin': 'almaz',
            #                             'objectId': 5565741,
            #                             'updatedAt': '2023-04-05T08:55:52.526Z'}],
            #                 'index': 0,
            #                 'pointCloudId': 19618685}],
            #     'framesCount': 1,
            #     'objects': [{'classId': 683259,
            #                 'classTitle': 'Car',
            #                 'createdAt': '2023-04-05T08:55:52.384Z',
            #                 'datasetId': 62664,
            #                 'entityId': None,
            #                 'id': 5565737,
            #                 'labelerLogin': 'almaz',
            #                 'tags': [],
            #                 'updatedAt': '2023-04-05T08:55:52.384Z'}],
            #     'tags': []
            # }
        """

        info = self._api.pointcloud.get_info_by_id(pointcloud_id)
        return self._download(info.dataset_id, pointcloud_id)

    def append(
        self,
        pointcloud_id: int,
        ann: PointcloudAnnotation,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> None:
        """
        Loads an PointcloudAnnotation to a given point cloud ID in the API.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param ann: PointcloudAnnotation object.
        :type ann: PointcloudAnnotation
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pointcloud_id = 198704259
            api.pointcloud.annotation.append(pointcloud_id, pointcloud_ann)
        """

        info = self._api.pointcloud.get_info_by_id(pointcloud_id)

        new_objects = []
        for object_3d in ann.objects:
            if key_id_map is not None and key_id_map.get_object_id(object_3d.key()) is not None:
                # object already uploaded
                continue
            new_objects.append(object_3d)

        self._append(
            self._api.pointcloud.tag,
            self._api.pointcloud.object,
            self._api.pointcloud.figure,
            info.project_id,
            info.dataset_id,
            pointcloud_id,
            ann.tags,
            PointcloudObjectCollection(new_objects),
            ann.figures,
            key_id_map,
        )

    async def download_async(
        self,
        pointcloud_id: int,
        semaphore: Optional[asyncio.Semaphore] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> Dict:
        """
        Download information about PointcloudAnnotation by Point Cloud ID from API asynchronously.

        :param pointcloud_id: Point Cloud ID in Supervisely.
        :type pointcloud_id: int
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param progress_cb: Progress callback to track download progress.
        :type progress_cb: Union[tqdm, Callable], optional
        :return: Information about PointcloudAnnotation in json format
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pointcloud_id = 198702499
            loop = sly.utils.get_or_create_event_loop()
            ann_info = loop.run_until_complete(api.pointcloud.annotation.download_async(pointcloud_id))
        """
        return await self.download_bulk_async(
            pointcloud_ids=[pointcloud_id],
            semaphore=semaphore,
            progress_cb=progress_cb,
        )

    async def download_bulk_async(
        self,
        pointcloud_ids: List[int],
        semaphore: Optional[asyncio.Semaphore] = None,
        progress_cb: Optional[Union[tqdm, Callable]] = None,
    ) -> List[Dict]:
        """
        Download information about PointcloudAnnotation in bulk by Point Cloud IDs from API asynchronously.

        :param pointcloud_ids: Point Cloud IDs in Supervisely.
        :type pointcloud_ids: List[int]
        :param semaphore: Semaphore to limit the number of parallel downloads.
        :type semaphore: asyncio.Semaphore, optional
        :param progress_cb: Progress callback to track download progress.
        :type progress_cb: Union[tqdm, Callable], optional
        :return: Information about PointcloudAnnotations in json format
        :rtype: :class:`list`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pointcloud_ids = [198702499, 198702500, 198702501]
            loop = sly.utils.get_or_create_event_loop()
            ann_infos = loop.run_until_complete(api.pointcloud.annotation.download_bulk_async(pointcloud_ids))
        """
        if semaphore is None:
            semaphore = self._api.get_default_semaphore()

        async def fetch_with_semaphore(batch):
            async with semaphore:
                json_data = {self._entity_ids_str: batch}
                response = await self._api.post_async(
                    self._method_download_bulk,
                    json=json_data,
                )
                if progress_cb is not None:
                    progress_cb(len(batch))
                return response.json()

        tasks = [fetch_with_semaphore(batch) for batch in batched(pointcloud_ids)]
        responses = await asyncio.gather(*tasks)
        json_response = [item for response in responses for item in response]
        return json_response



from datetime import datetime
from os import path as osp
from pathlib import Path
from typing import List

import numpy as np

from supervisely import fs, logger
from supervisely.geometry.cuboid_3d import Cuboid3d, Vector3d

DIR_NAMES = [
    "CAM_BACK",
    "CAM_BACK_LEFT",
    "CAM_BACK_RIGHT",
    "CAM_FRONT",
    "CAM_FRONT_LEFT",
    "CAM_FRONT_RIGHT",
    "LIDAR_TOP",
    "RADAR_FRONT",
    "RADAR_FRONT_LEFT",
    "RADAR_FRONT_RIGHT",
    "RADAR_BACK_LEFT",
    "RADAR_BACK_RIGHT",
]

TABLE_NAMES = [
    "category",
    "attribute",
    "visibility",
    "instance",
    "sensor",
    "calibrated_sensor",
    "ego_pose",
    "log",
    "scene",
    "sample",
    "sample_data",
    "sample_annotation",
    "map",
]


class Sample:
    """
    A class to represent a sample from the NuScenes dataset.
    """

    def __init__(self, timestamp, lidar_path, anns, cam_data):
        self.timestamp = datetime.utcfromtimestamp(timestamp / 1e6).isoformat()
        self.lidar_path = lidar_path
        self.anns = anns
        self.cam_data = cam_data

    @staticmethod
    def generate_boxes(nuscenes, boxes):
        """
        Generate ground truth boxes for a given set of boxes.

        Yields:
            tuple: A tuple containing:
                - gt_box (np.ndarray): A numpy array representing the ground truth box with concatenated location,
                                       dimensions, and rotation.
                - name (str): The name of the object.
                - instance_token (str): The instance token associated with the box.
        """
        locs = np.array([b.center for b in boxes]).reshape(-1, 3)
        dims = np.array([b.wlh for b in boxes]).reshape(-1, 3)
        rots = np.array([b.orientation.yaw_pitch_roll[0] for b in boxes]).reshape(-1, 1)

        gt_boxes = np.concatenate([locs, dims, -rots - np.pi / 2], axis=1)
        names = np.array([b.name for b in boxes])
        instance_tokens = [nuscenes.get("sample_annotation", box.token) for box in boxes]

        yield from zip(gt_boxes, names, instance_tokens)

    def convert_lidar_to_supervisely(self):
        """
        Converts a LiDAR point cloud file to the Supervisely format and saves it as a .pcd file.

        Returns:
            str: The file path of the saved .pcd file.
        """
        import open3d as o3d  # pylint: disable=import-error

        bin_file = Path(self.lidar_path)
        save_path = str(bin_file.with_suffix(".pcd"))

        b = np.fromfile(bin_file, dtype=np.float32).reshape(-1, 5)
        points = b[:, 0:3]
        intensity = b[:, 3]
        ring_index = b[:, 4]
        intensity_fake_rgb = np.zeros((intensity.shape[0], 3))
        intensity_fake_rgb[:, 0] = (
            intensity  # red The intensity measures the reflectivity of the objects
        )
        intensity_fake_rgb[:, 1] = (
            ring_index  # green ring index is the index of the laser ranging from 0 to 31
        )
        try:
            pc = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(points))
            pc.colors = o3d.utility.Vector3dVector(intensity_fake_rgb)
            o3d.io.write_point_cloud(save_path, pc)
        except Exception as e:
            logger.warning(f"Error converting lidar to supervisely format: {e}")
        return save_path


class AnnotationObject:
    """
    A class to represent an annotation object in the NuScenes dataset.

    Attributes:
    -----------
    name : str
        The name of the annotation object.
    bbox : np.ndarray
        The bounding box coordinates.
    instance_token : str
        The instance token associated with the annotation object.
    parent_token : str
        The token of instance preceding the current object instance.
    category : str
        The class name of the annotation object.
    attributes : List[str]
        The attribute names associated with the annotation object.
    visibility : str
        The visibility level of the annotation object.
    """

    def __init__(
        self,
        name: str,
        bbox: np.ndarray,
        instance_token: str,
        parent_token: str,
        category: str,
        attributes: List[str],
        visibility: str,
    ):
        self.name = name
        self.bbox = bbox
        self.instance_token = instance_token
        self.parent_token = parent_token

        self.category = category
        self.attributes = attributes
        self.visibility = visibility

    def to_supervisely(self):
        box = self.convert_nuscenes_to_BEVBox3D()

        bbox = box.to_xyzwhlr()
        dim = bbox[[3, 5, 4]]
        pos = bbox[:3] + [0, 0, dim[1] / 2]
        yaw = bbox[-1]

        position = Vector3d(float(pos[0]), float(pos[1]), float(pos[2]))
        rotation = Vector3d(0, 0, float(-yaw))
        dimension = Vector3d(float(dim[0]), float(dim[2]), float(dim[1]))
        geometry = Cuboid3d(position, rotation, dimension)

        return geometry

    def convert_nuscenes_to_BEVBox3D(self):
        import open3d as o3d  # pylint: disable=import-error

        box = self.bbox
        center = [float(box[0]), float(box[1]), float(box[2])]
        size = [float(box[3]), float(box[5]), float(box[4])]
        ry = float(box[6])
        yaw = ry - np.pi
        yaw = yaw - np.floor(yaw / (2 * np.pi) + 0.5) * 2 * np.pi
        world_cam = None
        return o3d.ml.datasets.utils.BEVBox3D(center, size, yaw, self.name, -1.0, world_cam)


class CamData:
    """
    A class to represent camera data and perform transformations between different coordinate systems.

    Attributes:
    -----------
    name : str
        The name of the sensor.
    path : str
        The path to the image file.
    imsize : tuple
        The size of the image (width, height).
    extrinsic : np.ndarray
        The extrinsic matrix (4x4) representing the transformation from the lidar to the camera coordinate system.
    intrinsic : np.ndarray
        The intrinsic matrix (3x3) representing the camera's intrinsic parameters.
    """

    def __init__(self, nuscenes, sensor_name, sensor_token, cs_record, ego_record):
        from nuscenes.utils.data_classes import (  # pylint: disable=import-error
            transform_matrix,
        )
        from pyquaternion import Quaternion  # pylint: disable=import-error

        img_path, boxes, cam_intrinsic = nuscenes.get_sample_data(sensor_token)
        if not osp.exists(img_path):
            return None

        sd_record_cam = nuscenes.get("sample_data", sensor_token)
        cs_record_cam = nuscenes.get("calibrated_sensor", sd_record_cam["calibrated_sensor_token"])
        ego_record_cam = nuscenes.get("ego_pose", sd_record_cam["ego_pose_token"])
        lid_to_ego = transform_matrix(
            cs_record["translation"],
            Quaternion(cs_record["rotation"]),
            inverse=False,
        )
        lid_ego_to_world = transform_matrix(
            ego_record["translation"],
            Quaternion(ego_record["rotation"]),
            inverse=False,
        )
        world_to_cam_ego = transform_matrix(
            ego_record_cam["translation"],
            Quaternion(ego_record_cam["rotation"]),
            inverse=True,
        )
        ego_to_cam = transform_matrix(
            cs_record_cam["translation"],
            Quaternion(cs_record_cam["rotation"]),
            inverse=True,
        )
        velo_to_cam = np.dot(
            ego_to_cam, np.dot(world_to_cam_ego, np.dot(lid_ego_to_world, lid_to_ego))
        )
        velo_to_cam_rot = velo_to_cam[:3, :3]
        velo_to_cam_trans = velo_to_cam[:3, 3]

        self.name = sensor_name
        self.path = str(img_path)
        self.imsize = (sd_record_cam["width"], sd_record_cam["height"])
        self.extrinsic = np.hstack((velo_to_cam_rot, velo_to_cam_trans.reshape(3, 1)))
        self.intrinsic = np.asarray(cs_record_cam["camera_intrinsic"])

    def get_info(self, timestamp):
        """
        Retrieves information about the image and its metadata.

        Args:
            timestamp (int): The timestamp associated with the image.

        Returns:
            tuple: A tuple containing the image path and a dictionary with image metadata.
        """
        sensors_to_skip = ["_intrinsic", "_extrinsic", "_imsize"]
        if not any([self.name.endswith(s) for s in sensors_to_skip]):
            image_name = fs.get_file_name_with_ext(self.path)
            sly_path_img = osp.join(osp.dirname(self.path), image_name)
            img_info = {
                "name": image_name,
                "meta": {
                    "deviceId": self.name,
                    "timestamp": timestamp,
                    "sensorsData": {
                        "extrinsicMatrix": list(self.extrinsic.flatten().astype(float)),
                        "intrinsicMatrix": list(self.intrinsic.flatten().astype(float)),
                    },
                },
            }
            return (sly_path_img, img_info)



from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, fields
from json import JSONDecodeError
from os.path import dirname, join
from typing import List

import requests

from supervisely import logger
from supervisely.api.api import Api, ApiField


@dataclass
class ExperimentInfo:
    experiment_name: str
    """Name of the experiment. Defined by the user in the training app"""
    framework_name: str
    """Name of the framework used in the experiment"""
    model_name: str
    """Name of the model used in the experiment. Defined by the user in the training app"""
    task_type: str
    """Task type of the experiment"""
    project_id: int
    """Project ID in Supervisely"""
    task_id: int
    """Task ID in Supervisely"""
    model_files: dict
    """Dictionary with paths to model files that needs to be downloaded for training"""
    checkpoints: List[str]
    """List of relative paths to checkpoints"""
    best_checkpoint: str
    """Name of the best checkpoint. Defined by the user in the training app"""
    export: dict
    """Dictionary with exported weights in different formats"""
    app_state: str
    """Path to file with settings that were used in the app"""
    model_meta: str
    """Path to file with model metadata such as model name, project id, project name and classes used for training"""
    train_val_split: str
    """Path to train and validation splits, which contains IDs of the images used in each split"""
    hyperparameters: str
    """Path to .yaml file with hyperparameters used in the experiment"""
    artifacts_dir: str
    """Path to the directory with artifacts"""
    datetime: str
    """Date and time when the experiment was started"""
    evaluation_report_id: int
    """ID of the evaluation report"""
    evaluation_metrics: dict
    """Evaluation metrics"""


def get_experiment_infos(
    api: Api, team_id: int, framework_name: str
) -> List[ExperimentInfo]:
    """
    Get experiments from the specified framework folder for Train v2

    :param api: Supervisely API client
    :type api: Api
    :param team_id: Team ID
    :type team_id: int
    :param framework_name: Name of the framework
    :type framework_name: str
    :return: List of ExperimentInfo objects
    :rtype: List[ExperimentInfo]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        api = sly.Api.from_env()
        team_id = sly.env.team_id()
        framework_name = "rt-detr"
        experiment_infos = sly.nn.training.experiments.get_experiment_infos(api, team_id, framework_name)
    """
    metadata_name = "experiment_info.json"
    experiments_folder = "/experiments"
    experiment_infos = []

    file_infos = api.file.list(
        team_id, experiments_folder, recursive=True, return_type="fileinfo"
    )
    sorted_experiment_paths = []
    for file_info in file_infos:
        if not file_info.path.endswith(metadata_name):
            continue

        experiment_dir = dirname(file_info.path)
        if experiment_dir.endswith(framework_name):
            experiment_path = join(experiment_dir, metadata_name)
            sorted_experiment_paths.append(experiment_path)

    def fetch_experiment_data(experiment_path: str):
        try:
            response = api.post(
                "file-storage.download",
                {ApiField.TEAM_ID: team_id, ApiField.PATH: experiment_path},
                stream=True,
            )
            response.raise_for_status()
            response_json = response.json()
            required_fields = {field.name for field in fields(ExperimentInfo)}
            missing_fields = required_fields - response_json.keys()
            if missing_fields:
                logger.debug(
                    f"Missing fields: {missing_fields} for '{experiment_path}'"
                )
                return None
            return ExperimentInfo(**response_json)
        except requests.exceptions.RequestException as e:
            logger.debug(f"Request failed for '{experiment_path}': {e}")
        except JSONDecodeError as e:
            logger.debug(f"JSON decode failed for '{experiment_path}': {e}")
        except TypeError as e:
            logger.error(f"TypeError for '{experiment_path}': {e}")
        return None

    with ThreadPoolExecutor() as executor:
        experiment_infos = list(
            executor.map(fetch_experiment_data, sorted_experiment_paths)
        )

    experiment_infos = [info for info in experiment_infos if info is not None]
    return experiment_infos



import random
from pathlib import Path
from typing import List

import supervisely.nn.benchmark.object_detection.text_templates as vis_texts
from supervisely._utils import batched
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagApplicableTo, TagMeta, TagValueType
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import ApiField
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.benchmark.base_visualizer import BaseVisualizer
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.object_detection.vis_metrics import (
    ConfidenceDistribution,
    ConfidenceScore,
    ConfusionMatrix,
    ExplorePredictions,
    F1ScoreAtDifferentIOU,
    FrequentlyConfused,
    IOUDistribution,
    KeyMetrics,
    ModelPredictions,
    OutcomeCounts,
    Overview,
    PerClassAvgPrecision,
    PerClassOutcomeCounts,
    PRCurve,
    PRCurveByClass,
    Precision,
    Recall,
    RecallVsPrecision,
    ReliabilityDiagram,
    Speedtest,
)
from supervisely.nn.benchmark.visualization.widgets import (
    ContainerWidget,
    MarkdownWidget,
    SidebarWidget,
)
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class ObjectDetectionVisualizer(BaseVisualizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vis_texts = vis_texts
        self._widgets = False
        self.ann_opacity = 0.4

        diff_project_info, diff_dataset_infos, _ = self._get_or_create_diff_project()
        self.eval_result.diff_project_info = diff_project_info
        self.eval_result.diff_dataset_infos = diff_dataset_infos
        self.eval_result.matched_pair_data = {}

        self.gt_project_path = str(Path(self.workdir).parent / "gt_project")
        self.pred_project_path = str(Path(self.workdir).parent / "pred_project")
        self.update_diff_annotations()

        # set filtered project meta
        self.eval_result.filtered_project_meta = self._get_filtered_project_meta(self.eval_result)
        self._get_sample_data_for_gallery()

    @property
    def cv_task(self):
        return CVTask.OBJECT_DETECTION

    def _create_widgets(self):
        # get cv task
        # Modal Gellery
        self.diff_modal = self._create_diff_modal_table()
        self.explore_modal = self._create_explore_modal_table(
            click_gallery_id=self.diff_modal.id, hover_text="Compare with GT"
        )

        # Notifcation
        self.clickable_label = self._create_clickable_label()

        # Overview
        me = self.api.user.get_my_info()
        overview = Overview(self.vis_texts, self.eval_result)
        self.header = overview.get_header(me.login)
        self.overview_md = overview.md

        # Key Metrics
        key_metrics = KeyMetrics(self.vis_texts, self.eval_result)
        self.key_metrics_md = key_metrics.md
        self.key_metrics_table = key_metrics.table
        self.overview_chart = key_metrics.chart

        # Explore Predictions
        explore_predictions = ExplorePredictions(
            self.vis_texts, self.eval_result, self.explore_modal, self.diff_modal
        )
        self.explore_predictions_md = explore_predictions.md
        self.explore_predictions_gallery = explore_predictions.gallery(opacity=self.ann_opacity)

        # Model Predictions
        model_predictions = ModelPredictions(self.vis_texts, self.eval_result, self.diff_modal)
        self.model_predictions_md = model_predictions.md
        self.model_predictions_table = model_predictions.table

        # Outcome Counts
        outcome_counts = OutcomeCounts(self.vis_texts, self.eval_result, self.explore_modal)
        self.outcome_counts_md = outcome_counts.md
        self.outcome_counts_chart = outcome_counts.chart

        # Recall
        recall = Recall(self.vis_texts, self.eval_result, self.explore_modal)
        self.recall_md = recall.md
        self.recall_notificaiton = recall.notification
        self.recall_per_class_md = recall.per_class_md
        self.recall_chart = recall.chart

        # Precision
        precision = Precision(self.vis_texts, self.eval_result, self.explore_modal)
        self.precision_md = precision.md
        self.precision_notification = precision.notification
        self.precision_per_class_md = precision.per_class_md
        self.precision_chart = precision.chart

        # RecallVsPrecision
        recall_vs_precision = RecallVsPrecision(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.recall_vs_precision_md = recall_vs_precision.md
        self.recall_vs_precision_chart = recall_vs_precision.chart

        # PRCurve
        pr_curve = PRCurve(self.vis_texts, self.eval_result)
        self.pr_curve_md = pr_curve.md
        self.pr_curve_notificaiton = pr_curve.notification
        self.pr_curve_chart = pr_curve.chart
        self.pr_curve_collapse = pr_curve.collapse

        # PRCurveByClass
        pr_curve_by_class = PRCurveByClass(self.vis_texts, self.eval_result, self.explore_modal)
        self.pr_curve_by_class_md = pr_curve_by_class.md
        self.pr_curve_by_class_chart = pr_curve_by_class.chart

        # ConfusionMatrix
        confusion_matrix = ConfusionMatrix(self.vis_texts, self.eval_result, self.explore_modal)
        self.confusion_matrix_md = confusion_matrix.md
        self.confusion_matrix_chart = confusion_matrix.chart

        # FrequentlyConfused
        frequently_confused = FrequentlyConfused(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.frequently_confused_present = frequently_confused.is_empty is False
        if self.frequently_confused_present:
            self.frequently_confused_md = frequently_confused.md
            self.frequently_confused_chart = frequently_confused.chart
        else:
            self.frequently_confused_md = frequently_confused.empty_md

        # IOUDistribution
        iou_distribution = IOUDistribution(self.vis_texts, self.eval_result)
        if self.cv_task in [CVTask.INSTANCE_SEGMENTATION, CVTask.SEMANTIC_SEGMENTATION]:
            iou_distribution.md_title = "Mask Accuracy (IoU)"
        self.iou_distribution_md = iou_distribution.md
        self.iou_distribution_md_iou_distribution = iou_distribution.md_iou_distribution
        self.iou_distribution_notification = iou_distribution.notification
        self.iou_distribution_chart = iou_distribution.chart

        # ReliabilityDiagram
        reliability_diagram = ReliabilityDiagram(self.vis_texts, self.eval_result)
        self.reliability_diagram_md_calibration_score = reliability_diagram.md_calibration_score
        self.reliability_diagram_collapse_1 = reliability_diagram.collapse_tip
        self.reliability_diagram_md_calibration_score_2 = reliability_diagram.md_calibration_score_2
        self.reliability_diagram_md_reliability_diagram = reliability_diagram.md_reliability_diagram
        self.reliability_diagram_notification = reliability_diagram.notification
        self.reliability_diagram_chart = reliability_diagram.chart
        self.reliability_diagram_collapse_2 = reliability_diagram.collapse

        # ConfidenceScore
        confidence_score = ConfidenceScore(self.vis_texts, self.eval_result)
        self.confidence_score_md_confidence_score = confidence_score.md_confidence_score
        self.confidence_score_notification = confidence_score.notification
        self.confidence_score_chart = confidence_score.chart
        self.confidence_score_md_confidence_score_2 = confidence_score.md_confidence_score_2
        self.confidence_score_collapse_conf_score = confidence_score.collapse_conf_score
        self.confidence_score_md_confidence_score_3 = confidence_score.md_confidence_score_3

        # F1ScoreAtDifferentIOU
        f1_score_at_different_iou = F1ScoreAtDifferentIOU(self.vis_texts, self.eval_result)
        self.f1_score_at_different_iou_md = f1_score_at_different_iou.md
        self.f1_score_at_different_iou_chart = f1_score_at_different_iou.chart

        # ConfidenceDistribution
        confidence_distribution = ConfidenceDistribution(self.vis_texts, self.eval_result)
        self.confidence_distribution_md = confidence_distribution.md
        self.confidence_distribution_chart = confidence_distribution.chart

        # PerClassAvgPrecision
        per_class_avg_precision = PerClassAvgPrecision(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.per_class_avg_precision_md = per_class_avg_precision.md
        self.per_class_avg_precision_chart = per_class_avg_precision.chart

        # PerClassOutcomeCounts
        per_class_outcome_counts = PerClassOutcomeCounts(
            self.vis_texts, self.eval_result, self.explore_modal
        )
        self.per_class_outcome_counts_md = per_class_outcome_counts.md
        self.per_class_outcome_counts_md_2 = per_class_outcome_counts.md_2
        self.per_class_outcome_counts_collapse = per_class_outcome_counts.collapse
        self.per_class_outcome_counts_chart = per_class_outcome_counts.chart

        # Speedtest init here for overview
        speedtest = Speedtest(self.vis_texts, self.eval_result)
        self.speedtest_present = not speedtest.is_empty()
        self.speedtest_batch_sizes_cnt = speedtest.num_batche_sizes
        if self.speedtest_present:
            self.speedtest_md_intro = speedtest.intro_md
            self.speedtest_table_md = speedtest.table_md
            self.speedtest_table = speedtest.table
            if self.speedtest_batch_sizes_cnt > 1:
                self.speedtest_chart_md = speedtest.chart_md
                self.speedtest_chart = speedtest.chart

        self._widgets = True

    def _create_layout(self):
        if not self._widgets:
            self._create_widgets()

        is_anchors_widgets = [
            # Overview
            (0, self.header),
            (1, self.overview_md),
            # KeyMetrics
            (1, self.key_metrics_md),
            (0, self.key_metrics_table),
            (0, self.overview_chart),
            # ExplorePredictions
            (1, self.explore_predictions_md),
            (0, self.explore_predictions_gallery),
            # ModelPredictions
            (1, self.model_predictions_md),
            (0, self.model_predictions_table),
            # OutcomeCounts
            (1, self.outcome_counts_md),
            (0, self.clickable_label),
            (0, self.outcome_counts_chart),
            # Recall
            (1, self.recall_md),
            (0, self.recall_notificaiton),
            (0, self.recall_per_class_md),
            (0, self.clickable_label),
            (0, self.recall_chart),
            # Precision
            (1, self.precision_md),
            (0, self.precision_notification),
            (0, self.precision_per_class_md),
            (0, self.clickable_label),
            (0, self.precision_chart),
            # RecallVsPrecision
            (1, self.recall_vs_precision_md),
            (0, self.clickable_label),
            (0, self.recall_vs_precision_chart),
            # PRCurve
            (1, self.pr_curve_md),
            (0, self.pr_curve_notificaiton),
            (0, self.pr_curve_chart),
            (0, self.pr_curve_collapse),
            # PRCurveByClass
            (0, self.pr_curve_by_class_md),
            (0, self.clickable_label),
            (0, self.pr_curve_by_class_chart),
            # ConfusionMatrix
            (1, self.confusion_matrix_md),
            (0, self.clickable_label),
            (0, self.confusion_matrix_chart),
            # FrequentlyConfused
            (1, self.frequently_confused_md),
        ]
        if self.frequently_confused_present:
            is_anchors_widgets.append((0, self.clickable_label))
            is_anchors_widgets.append((0, self.frequently_confused_chart))

        is_anchors_widgets.extend(
            [
                # IOUDistribution
                (1, self.iou_distribution_md),
                (0, self.iou_distribution_md_iou_distribution),
                (0, self.iou_distribution_notification),
                (0, self.iou_distribution_chart),
                # ReliabilityDiagram
                (1, self.reliability_diagram_md_calibration_score),
                (0, self.reliability_diagram_collapse_1),
                (0, self.reliability_diagram_md_calibration_score_2),
                (1, self.reliability_diagram_md_reliability_diagram),
                (0, self.reliability_diagram_notification),
                (0, self.reliability_diagram_chart),
                (0, self.reliability_diagram_collapse_2),
                # ConfidenceScore
                (1, self.confidence_score_md_confidence_score),
                (0, self.confidence_score_notification),
                (0, self.confidence_score_chart),
                (0, self.confidence_score_md_confidence_score_2),
                (0, self.confidence_score_collapse_conf_score),
                (0, self.confidence_score_md_confidence_score_3),
                # F1ScoreAtDifferentIOU
                (1, self.f1_score_at_different_iou_md),
                (0, self.f1_score_at_different_iou_chart),
                # ConfidenceDistribution
                (1, self.confidence_distribution_md),
                (0, self.confidence_distribution_chart),
                # PerClassAvgPrecision
                (1, self.per_class_avg_precision_md),
                (0, self.clickable_label),
                (0, self.per_class_avg_precision_chart),
                # PerClassOutcomeCounts
                (1, self.per_class_outcome_counts_md),
                (0, self.per_class_outcome_counts_md_2),
                (0, self.per_class_outcome_counts_collapse),
                (0, self.clickable_label),
                (0, self.per_class_outcome_counts_chart),
            ]
        )

        if self.speedtest_present:
            # SpeedTest
            is_anchors_widgets.append((1, self.speedtest_md_intro))
            is_anchors_widgets.append((0, self.speedtest_table_md))
            is_anchors_widgets.append((0, self.speedtest_table))
            if self.speedtest_batch_sizes_cnt > 1:
                is_anchors_widgets.append((0, self.speedtest_chart_md))
                is_anchors_widgets.append((0, self.speedtest_chart))
        anchors = []
        for is_anchor, widget in is_anchors_widgets:
            if is_anchor:
                anchors.append(widget.id)

        sidebar = SidebarWidget(widgets=[i[1] for i in is_anchors_widgets], anchors=anchors)
        layout = ContainerWidget(
            widgets=[sidebar, self.explore_modal, self.diff_modal],
            name="main_container",
        )
        return layout

    def _create_clickable_label(self):
        return MarkdownWidget(name="clickable_label", title="", text=self.vis_texts.clickable_label)

    def update_diff_annotations(self):
        pred_project_id = self.eval_result.pred_project_id
        pred_project_meta = self.eval_result.pred_project_meta
        meta = self._update_pred_meta_with_tags(pred_project_id, pred_project_meta)
        self.eval_result.pred_project_meta = meta

        self._update_diff_meta(meta)

        self._add_tags_to_pred_project(
            self.eval_result.mp.matches, self.eval_result.pred_project_id
        )

        gt_project = Project(self.gt_project_path, OpenMode.READ)
        pred_project = Project(self.pred_project_path, OpenMode.READ)
        diff_dataset_id_map = {ds.id: ds for ds in self.eval_result.diff_dataset_infos}
        logger.info(f"Diff datasets names: {[ds.name for ds in diff_dataset_id_map.values()]}")

        def _get_full_name(ds_id: int):
            ds_info = diff_dataset_id_map[ds_id]
            if ds_info.parent_id is None:
                return ds_info.name
            return f"{_get_full_name(ds_info.parent_id)}/{ds_info.name}"

        diff_dataset_name_map = {_get_full_name(i): ds for i, ds in diff_dataset_id_map.items()}

        matched_id_map = self._get_matched_id_map()  # dt_id -> gt_id
        matched_gt_ids = set(matched_id_map.values())

        outcome_tag = meta.get_tag_meta("outcome")
        conf_meta = meta.get_tag_meta("confidence")
        if conf_meta is None:
            conf_meta = meta.get_tag_meta("conf")
        match_tag = meta.get_tag_meta("matched_gt_id")

        pred_tag_list = []
        with self.pbar(
            message="Visualizations: Creating difference project", total=pred_project.total_items
        ) as progress:
            logger.debug(
                "Creating diff project data",
                extra={
                    "pred_project": [ds.name for ds in pred_project.datasets],
                    "gt_project": [ds.name for ds in gt_project.datasets],
                },
            )
            for pred_dataset in pred_project.datasets:
                pred_dataset: Dataset
                gt_dataset: Dataset = gt_project.datasets.get(pred_dataset.name)
                diff_dataset_info = diff_dataset_name_map[pred_dataset.name]
                for batch_names in batched(pred_dataset.get_items_names(), 100):
                    diff_anns = []
                    gt_image_ids = []
                    pred_img_ids = []
                    for item_name in batch_names:
                        gt_image_info = gt_dataset.get_image_info(item_name)
                        gt_image_ids.append(gt_image_info.id)
                        pred_image_info = pred_dataset.get_image_info(item_name)
                        pred_img_ids.append(pred_image_info.id)
                        gt_ann = gt_dataset.get_ann(item_name, gt_project.meta)
                        pred_ann = pred_dataset.get_ann(item_name, pred_project.meta)
                        labels = []

                        # TP and FP
                        for label in pred_ann.labels:
                            match_tag_id = matched_id_map.get(label.geometry.sly_id)
                            value = "TP" if match_tag_id else "FP"
                            pred_tag_list.append(
                                {
                                    "tagId": outcome_tag.sly_id,
                                    "figureId": label.geometry.sly_id,
                                    "value": value,
                                }
                            )
                            conf = 1
                            for tag in label.tags.items():
                                tag: Tag
                                if tag.name in ["confidence", "conf"]:
                                    conf = tag.value
                                    break

                            if conf < self.eval_result.mp.f1_optimal_conf:
                                continue  # do not add labels with low confidence to diff project
                            if match_tag_id:
                                continue  # do not add TP labels to diff project
                            label = label.add_tag(Tag(outcome_tag, value))
                            label = label.add_tag(Tag(match_tag, int(label.geometry.sly_id)))
                            labels.append(label)

                        # FN
                        for label in gt_ann.labels:
                            if self.eval_result.classes_whitelist:
                                if label.obj_class.name not in self.eval_result.classes_whitelist:
                                    continue
                            if label.geometry.sly_id not in matched_gt_ids:
                                if self._is_label_compatible_to_cv_task(label):
                                    new_label = label.add_tags(
                                        [Tag(outcome_tag, "FN"), Tag(conf_meta, 1)]
                                    )
                                    labels.append(new_label)

                        diff_ann = Annotation(gt_ann.img_size, labels)
                        diff_anns.append(diff_ann)

                        # comparison data
                        self._update_match_data(
                            gt_image_info.id,
                            gt_image_info=gt_image_info,
                            pred_image_info=pred_image_info,
                            gt_annotation=gt_ann,
                            pred_annotation=pred_ann,
                            diff_annotation=diff_ann,
                        )

                    diff_img_infos = self.api.image.copy_batch(diff_dataset_info.id, pred_img_ids)
                    ids = [img.id for img in diff_img_infos]
                    self.api.annotation.upload_anns(ids, diff_anns, progress_cb=progress.update)
                    for gt_img_id, diff_img_info in zip(gt_image_ids, diff_img_infos):
                        self._update_match_data(gt_img_id, diff_image_info=diff_img_info)

        with self.pbar(
            message="Visualizations: Append tags to predictions", total=len(pred_tag_list)
        ) as p:
            self.api.image.tag.add_to_objects(
                self.eval_result.pred_project_id, pred_tag_list, progress=p
            )

    def _init_match_data(self):
        gt_project = Project(self.gt_project_path, OpenMode.READ)
        pred_project = Project(self.pred_project_path, OpenMode.READ)
        diff_dataset_id_map = {ds.id: ds for ds in self.eval_result.diff_dataset_infos}
        logger.info(f"Diff datasets names: {[ds.name for ds in diff_dataset_id_map.values()]}")

        def _get_full_name(ds_id: int):
            ds_info = diff_dataset_id_map[ds_id]
            if ds_info.parent_id is None:
                return ds_info.name
            return f"{_get_full_name(ds_info.parent_id)}/{ds_info.name}"

        diff_dataset_name_map = {_get_full_name(i): ds for i, ds in diff_dataset_id_map.items()}

        meta_json = self.api.project.get_meta(self.eval_result.diff_project_info.id)
        self.eval_result.diff_project_meta = ProjectMeta.from_json(meta_json)

        with self.pbar(
            message="Visualizations: Initializing match data", total=pred_project.total_items
        ) as p:
            for pred_dataset in pred_project.datasets:
                pred_dataset: Dataset
                gt_dataset: Dataset = gt_project.datasets.get(pred_dataset.name)
                try:
                    diff_dataset_info = diff_dataset_name_map[pred_dataset.name]
                except KeyError:
                    raise RuntimeError(
                        f"Difference project was not created properly. Dataset {pred_dataset.name} is missing"
                    )

                for item_names_batch in batched(pred_dataset.get_items_names(), 50):
                    # diff project may be not created yet
                    item_names_batch.sort()
                    try:
                        diff_img_infos_batch: List[ImageInfo] = sorted(
                            self.api.image.get_list(
                                diff_dataset_info.id,
                                filters=[
                                    {
                                        ApiField.FIELD: ApiField.NAME,
                                        ApiField.OPERATOR: "in",
                                        ApiField.VALUE: item_names_batch,
                                    }
                                ],
                                force_metadata_for_links=False,
                            ),
                            key=lambda x: x.name,
                        )
                        diff_anns_batch_dict = {
                            ann_info.image_id: Annotation.from_json(
                                ann_info.annotation, self.eval_result.diff_project_meta
                            )
                            for ann_info in self.api.annotation.download_batch(
                                diff_dataset_info.id,
                                [img_info.id for img_info in diff_img_infos_batch],
                                force_metadata_for_links=False,
                            )
                        }
                        assert (
                            len(item_names_batch)
                            == len(diff_img_infos_batch)
                            == len(diff_anns_batch_dict)
                        ), "Some images are missing in the difference project"

                        for item_name, diff_img_info in zip(item_names_batch, diff_img_infos_batch):
                            assert (
                                item_name == diff_img_info.name
                            ), "Image names in difference project and prediction project do not match"
                            gt_image_info = gt_dataset.get_image_info(item_name)
                            pred_image_info = pred_dataset.get_image_info(item_name)
                            gt_ann = gt_dataset.get_ann(item_name, gt_project.meta)
                            pred_ann = pred_dataset.get_ann(item_name, pred_project.meta)
                            diff_ann = diff_anns_batch_dict[diff_img_info.id]

                            self._update_match_data(
                                gt_image_info.id,
                                gt_image_info=gt_image_info,
                                pred_image_info=pred_image_info,
                                diff_image_info=diff_img_info,
                                gt_annotation=gt_ann,
                                pred_annotation=pred_ann,
                                diff_annotation=diff_ann,
                            )

                        p.update(len(item_names_batch))
                    except Exception:
                        raise RuntimeError("Difference project was not created properly")

    def _update_pred_meta_with_tags(self, project_id: int, meta: ProjectMeta) -> ProjectMeta:
        old_meta = meta
        outcome_tag = TagMeta(
            "outcome",
            value_type=TagValueType.ONEOF_STRING,
            possible_values=["TP", "FP", "FN"],
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        match_tag = TagMeta(
            "matched_gt_id",
            TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        iou_tag = TagMeta(
            "iou",
            TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )
        confidence_tag = TagMeta(
            "confidence",
            value_type=TagValueType.ANY_NUMBER,
            applicable_to=TagApplicableTo.OBJECTS_ONLY,
        )

        for tag in [outcome_tag, match_tag, iou_tag]:
            if meta.get_tag_meta(tag.name) is None:
                meta = meta.add_tag_meta(tag)

        if meta.get_tag_meta("confidence") is None and meta.get_tag_meta("conf") is None:
            meta = meta.add_tag_meta(confidence_tag)

        if old_meta == meta:
            return meta

        meta = self.api.project.update_meta(project_id, meta)
        return meta

    def _update_diff_meta(self, meta: ProjectMeta):
        new_obj_classes = []
        for obj_class in meta.obj_classes:
            new_obj_classes.append(obj_class.clone(geometry_type=AnyGeometry))
        meta = meta.clone(obj_classes=new_obj_classes)
        self.eval_result.diff_project_meta = self.api.project.update_meta(
            self.eval_result.diff_project_info.id, meta
        )

    def _add_tags_to_pred_project(self, matches: list, pred_project_id: int):

        # get tag metas
        # outcome_tag_meta = self.dt_project_meta.get_tag_meta("outcome")
        match_tag_meta = self.eval_result.pred_project_meta.get_tag_meta("matched_gt_id")
        iou_tag_meta = self.eval_result.pred_project_meta.get_tag_meta("iou")

        # mappings
        gt_ann_mapping = self.eval_result.click_data.gt_id_mapper.map_obj
        dt_ann_mapping = self.eval_result.click_data.dt_id_mapper.map_obj

        # add tags to objects
        logger.info("Adding tags to DT project")

        with self.pbar(
            message="Visualizations: Adding tags to predictions", total=len(matches)
        ) as p:
            for batch in batched(matches, 100):
                pred_tag_list = []
                for match in batch:
                    if match["type"] == "TP":
                        outcome = "TP"
                        matched_gt_id = gt_ann_mapping[match["gt_id"]]
                        ann_dt_id = dt_ann_mapping[match["dt_id"]]
                        iou = match["iou"]
                        # api.advanced.add_tag_to_object(outcome_tag_meta.sly_id, ann_dt_id, str(outcome))
                        if matched_gt_id is not None:
                            pred_tag_list.extend(
                                [
                                    {
                                        "tagId": match_tag_meta.sly_id,
                                        "figureId": ann_dt_id,
                                        "value": int(matched_gt_id),
                                    },
                                    {
                                        "tagId": iou_tag_meta.sly_id,
                                        "figureId": ann_dt_id,
                                        "value": float(iou),
                                    },
                                ]
                            )
                        else:
                            continue
                    elif match["type"] == "FP":
                        outcome = "FP"
                        # api.advanced.add_tag_to_object(outcome_tag_meta.sly_id, ann_dt_id, str(outcome))
                    elif match["type"] == "FN":
                        outcome = "FN"
                    else:
                        raise ValueError(f"Unknown match type: {match['type']}")

                self.api.image.tag.add_to_objects(pred_project_id, pred_tag_list)
                p.update(len(batch))

    def _get_matched_id_map(self):
        gt_ann_mapping = self.eval_result.click_data.gt_id_mapper.map_obj
        dt_ann_mapping = self.eval_result.click_data.dt_id_mapper.map_obj
        dtId2matched_gt_id = {}
        for match in self.eval_result.mp.matches_filtered:
            if match["type"] == "TP":
                dtId2matched_gt_id[dt_ann_mapping[match["dt_id"]]] = gt_ann_mapping[match["gt_id"]]
        return dtId2matched_gt_id

    def _is_label_compatible_to_cv_task(self, label: Label):
        if self.cv_task == CVTask.OBJECT_DETECTION:
            return isinstance(label.geometry, Rectangle)
        elif self.cv_task == CVTask.INSTANCE_SEGMENTATION:
            return isinstance(label.geometry, (Bitmap, Polygon))
        elif self.cv_task == CVTask.SEMANTIC_SEGMENTATION:
            return isinstance(label.geometry, Bitmap)
        return False

    def _get_sample_data_for_gallery(self):
        # get sample images with annotations for visualization (Prediction project)
        pred_ds = random.choice(self.eval_result.pred_dataset_infos)
        self.eval_result.sample_images = self.api.image.get_list(pred_ds.id, limit=9)
        image_ids = [x.id for x in self.eval_result.sample_images]
        self.eval_result.sample_anns = self.api.annotation.download_batch(pred_ds.id, image_ids)



import unittest

from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag import Tag, TagValueType
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection
from supervisely.geometry.rectangle import Rectangle
from supervisely.metric.map_metric import MAPMetric, AP
from supervisely.project.project_meta import ProjectMeta

class PointTest(unittest.TestCase):
    def setUp(self):
        self._obj_class_gt = ObjClass(name='a', geometry_type=Rectangle)
        self._obj_class_pred = ObjClass(name='b', geometry_type=Rectangle)
        self._confidence_tag_meta = TagMeta(name='confidence', value_type=TagValueType.ANY_NUMBER)
        self._meta = ProjectMeta(
            obj_classes=ObjClassCollection([self._obj_class_gt, self._obj_class_pred]),
            tag_metas=TagMetaCollection([self._confidence_tag_meta]))

        # Will match self._pred_obj_1
        self._gt_obj_1 = Label(obj_class=self._obj_class_gt, geometry=Rectangle(0, 0, 10, 10))

        # Will match self._pred_obj_3
        self._gt_obj_2 = Label(obj_class=self._obj_class_gt, geometry=Rectangle(13, 13, 15, 15))

        # Will be a false negative
        self._gt_obj_3 = Label(obj_class=self._obj_class_gt, geometry=Rectangle(43, 43, 45, 45))

        # Will match self._gt_obj_1
        self._pred_obj_1 = Label(
            obj_class=self._obj_class_pred,
            geometry=Rectangle(0, 0, 9, 9),
            tags=TagCollection([Tag(meta=self._confidence_tag_meta, value=0.7)]))

        # Will be a false positive (self._pred_obj_1 has higher IoU).
        self._pred_obj_2 = Label(
            obj_class=self._obj_class_pred,
            geometry=Rectangle(0, 0, 8, 8),
            tags=TagCollection([Tag(meta=self._confidence_tag_meta, value=0.6)]))

        # Will match self._gt_obj_2
        self._pred_obj_3 = Label(
            obj_class=self._obj_class_pred,
            geometry=Rectangle(13, 13, 15, 15),
            tags=TagCollection([Tag(meta=self._confidence_tag_meta, value=0.1)]))

        # More false positives.
        self._pred_objs_fp = [
            Label(obj_class=self._obj_class_pred,
                  geometry=Rectangle(20, 20, 30, 30),
                  tags=TagCollection([Tag(meta=self._confidence_tag_meta, value=v / 100)]))
            for v in range(15, 85, 10)]

        self._metric_calculator = MAPMetric(class_mapping={'a': 'b'}, iou_threshold=0.5)

    def test_empty_gt(self):
        ann = Annotation(
            img_size=[100, 100], labels=[self._pred_obj_1, self._pred_obj_2, self._pred_obj_3] + self._pred_objs_fp)
        self._metric_calculator.add_pair(ann, ann)
        self.assertEqual(self._metric_calculator.get_total_metrics()[AP], 0)

    def test_empty_predictions(self):
        ann = Annotation(
            img_size=[100, 100], labels=[self._gt_obj_1, self._gt_obj_2, self._gt_obj_3])
        self._metric_calculator.add_pair(ann, ann)
        self.assertEqual(self._metric_calculator.get_total_metrics()[AP], 0)

    def test_with_matches(self):
        ann = Annotation(
            img_size=[100, 100],
            labels=[self._gt_obj_1, self._gt_obj_2, self._gt_obj_3,
                    self._pred_obj_1, self._pred_obj_2, self._pred_obj_3] + self._pred_objs_fp)
        self._metric_calculator.add_pair(ann, ann)

        # Sorted matches by confidence:
        # 0.75 - recall 0   precision 0
        # 0.7  + recall 1/3 precision 1/2
        # 0.65 - recall 1/3 precision 1/3
        # 0.6  - recall 1/3 precision 1/4
        # 0.55 - recall 1/3 precision 1/5
        # 0.45 - recall 1/3 precision 1/6
        # 0.35 - recall 1/3 precision 1/7
        # 0.25 - recall 1/3 precision 1/8
        # 0.15 - recall 2/3 precision 1/9
        # 0.1  + recall 2/3 precision 2/10

        # Recalls 0.7, 0.8, 0.9, 1.0 -> max precision 0.
        # Recalls 0.6, 0.5, 0.4      -> max precision 2/10
        # Recalls 0.3, 0.2, 0.1, 0.0 -> max precision 1/2
        expected_map = (4 * 0.0 + 3 * (2/10) + 4 * 1/2) / 11
        self.assertEqual(self._metric_calculator.get_total_metrics()[AP], expected_map)


if __name__ == '__main__':
    unittest.main()



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class DialogWindowBase(Exception):
    def __init__(self, title, description, status):
        self.title = title
        self.description = description
        self.status = status
        super().__init__(self.get_message())

    def get_message(self):
        return f"{self.title}: {self.description}"

    def __str__(self):
        return self.get_message()


class DialogWindowError(DialogWindowBase):
    def __init__(self, title, description):
        super().__init__(title, description, "error")


class DialogWindowWarning(DialogWindowBase):
    def __init__(self, title, description):
        super().__init__(title, description, "warning")


# for compatibility
class DialogWindowMessage(DialogWindowError):
    def __init__(self, title, description):
        super().__init__(title, description, "info")  # pylint: disable=too-many-function-args


def show_dialog(
    title, description, status: Literal["info", "success", "warning", "error"] = "info"
):
    from supervisely.app import StateJson, DataJson

    StateJson()["slyAppShowDialog"] = True
    DataJson()["slyAppDialogTitle"] = title
    DataJson()["slyAppDialogMessage"] = description
    DataJson()["slyAppDialogStatus"] = status
    DataJson().send_changes()
    StateJson().send_changes()



from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class MMClassification(BaseTrainArtifacts):
    def __init__(self, team_id: int):
        super().__init__(team_id)

        self._app_name = "Train MMClassification"
        self._framework_name = "MMClassification"
        self._framework_folder = "/mmclassification"
        self._weights_folder = "checkpoints"
        self._task_type = "classification"
        self._weights_ext = ".pth"
        self._pattern = re_compile(r"^/mmclassification/\d+_[^/]+/?$")
        self._available_task_types: List[str] = ["classification"]

    def get_task_id(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        session_id, _ = parts[2].split("_", 1)
        return session_id

    def get_project_name(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        _, project_name = parts[2].split("_", 1)
        return project_name

    def get_task_type(self, artifacts_folder: str) -> str:
        return self._task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return None



import os
import shutil
from pathlib import Path
from typing import Callable, List, Literal, Optional, Tuple, Union

import yaml
from tqdm import tqdm

from supervisely._utils import generate_free_name
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.geometry.alpha_mask import AlphaMask
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes, KeypointsTemplate, Node
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.color import generate_rgb
from supervisely.io.fs import get_file_name_with_ext, touch
from supervisely.project.project import Dataset, OpenMode, Project
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

YOLO_DETECTION_COORDS_NUM = 4
YOLO_SEGM_MIN_COORDS_NUM = 6
YOLO_KEYPOINTS_MIN_COORDS_NUM = 6

coco_classes = [
    "person",
    "bicycle",
    "car",
    "motorcycle",
    "airplane",
    "bus",
    "train",
    "truck",
    "boat",
    "traffic light",
    "fire hydrant",
    "stop sign",
    "parking meter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "backpack",
    "umbrella",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sports ball",
    "kite",
    "baseball bat",
    "baseball glove",
    "skateboard",
    "surfboard",
    "tennis racket",
    "bottle",
    "wine glass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hot dog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "couch",
    "potted plant",
    "bed",
    "dining table",
    "toilet",
    "tv",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cell phone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddy bear",
    "hair drier",
    "toothbrush",
]


def generate_colors(count: int) -> List[List[int]]:
    colors = []
    for _ in range(count):
        new_color = generate_rgb(colors)
        colors.append(new_color)
    return colors


def get_coordinates(line: str) -> Tuple[int, List[float]]:
    """
    Parse coordinates from a line in the YOLO format.
    """
    class_index = int(line[0])
    coords = list(map(float, line[1:]))
    return class_index, coords


def convert_rectangle(
    img_height: int,
    img_width: int,
    coords: List[float],
    **kwargs,
) -> Rectangle:
    """
    Convert rectangle coordinates from relative (0-1) to absolute (px) values.
    """
    x_center, y_center, ann_width, ann_height = coords
    x_center = float(x_center)
    y_center = float(y_center)
    ann_width = float(ann_width)
    ann_height = float(ann_height)

    px_x_center = x_center * img_width
    px_y_center = y_center * img_height

    px_ann_width = ann_width * img_width
    px_ann_height = ann_height * img_height

    left = int(px_x_center - (px_ann_width / 2))
    right = int(px_x_center + (px_ann_width / 2))

    top = int(px_y_center - (px_ann_height / 2))
    bottom = int(px_y_center + (px_ann_height / 2))

    # check if the coordinates are within the image
    left, top = max(0, left), max(0, top)
    right, bottom = min(img_width, right), min(img_height, bottom)

    return Rectangle(top, left, bottom, right)


def validate_polygon_coords(coords: List[float]) -> List[float]:
    """
    Check and correct polygon coordinates:
    - remove the last point if it is the same as the first one
    """
    if coords[0] == coords[-2] and coords[1] == coords[-1]:
        return coords[:-2]
    return coords


def convert_polygon(
    img_height: int,
    img_width: int,
    coords: List[float],
    **kwargs,
) -> Union[Polygon, None]:
    """
    Convert polygon coordinates from relative (0-1) to absolute (px) values.
    """
    coords = validate_polygon_coords(coords)
    if len(coords) < 6:
        logger.warning("Polygon has less than 3 points. Skipping.")
        return None

    exterior = []
    for i in range(0, len(coords), 2):
        x = coords[i]
        y = coords[i + 1]
        px_x = min(img_width, max(0, int(x * img_width)))
        px_y = min(img_height, max(0, int(y * img_height)))
        exterior.append([px_y, px_x])
    return Polygon(exterior=exterior)


def convert_keypoints(
    img_height: int,
    img_width: int,
    num_keypoints: int,
    num_dims: int,
    coords: List[float],
    **kwargs,
) -> Union[GraphNodes, None]:
    """
    Convert keypoints coordinates from relative (0-1) to absolute (px) values.
    """
    nodes = []
    step = 3 if num_dims == 3 else 2
    shift = 4
    for i in range(shift, num_keypoints * step + shift, step):
        x = coords[i]
        y = coords[i + 1]
        visibility = int(coords[i + 2]) if num_dims == 3 else 2
        if visibility in [0, 1]:
            continue  # skip invisible keypoints
        px_x = min(img_width, max(0, int(x * img_width)))
        px_y = min(img_height, max(0, int(y * img_height)))
        node = Node(row=px_y, col=px_x)  # , disabled=v)
        nodes.append(node)
    if len(nodes) > 0:
        return GraphNodes(nodes)


def create_geometry_config(num_keypoints: int = None) -> KeypointsTemplate:
    """
    Create a template for keypoints with the specified number of keypoints.
    """
    i, j = 0, 0
    template = KeypointsTemplate()
    for p in list(range(num_keypoints)):
        template.add_point(label=str(p), row=i, col=j)
        j += 1
        i += 1

    return template


def is_applicable_for_rectangles(coords: List[float], **kwargs) -> bool:
    """
    Check if the coordinates are applicable for rectangles.
    """
    return len(coords) == YOLO_DETECTION_COORDS_NUM


def is_applicable_for_polygons(
    with_keypoint: bool,
    coords: List[float],
    **kwargs,
) -> bool:
    """
    Check if the coordinates are applicable for polygons.

    :param with_keypoint: Whether the YAML config file contains keypoints.
    :type with_keypoint: bool
    """
    if with_keypoint:
        return False
    return len(coords) >= YOLO_SEGM_MIN_COORDS_NUM and len(coords) % 2 == 0


def is_applicable_for_keypoints(
    with_keypoint: bool,
    num_keypoints: int,
    num_dims: int,
    coords: List[float],
    **kwargs,
) -> bool:
    """
    Check if the coordinates are applicable for keypoints.
    """
    if not with_keypoint or not num_keypoints or not num_dims:
        return False
    if len(coords) < YOLO_KEYPOINTS_MIN_COORDS_NUM:
        return False
    return len(coords) == num_keypoints * num_dims + 4


APPLICABLE_GEOMETRIES_MAP = {
    Rectangle: is_applicable_for_rectangles,
    Polygon: is_applicable_for_polygons,
    GraphNodes: is_applicable_for_keypoints,
}


def detect_geometry(
    coords: List[float],
    with_keypoint: bool,
    num_keypoints: int,
    num_dims: int,
) -> Union[Rectangle, Polygon, GraphNodes, None]:
    """
    Detect the geometry type based on the coordinates and the configuration.
    """
    for geometry, is_applicable in APPLICABLE_GEOMETRIES_MAP.items():
        if is_applicable(
            with_keypoint=with_keypoint,
            num_keypoints=num_keypoints,
            num_dims=num_dims,
            coords=coords,
        ):
            return geometry


GEOMETRY_CONVERTERS = {
    Rectangle: convert_rectangle,
    Polygon: convert_polygon,
    GraphNodes: convert_keypoints,
}


def get_geometry(
    geometry_type: Union[Rectangle, Polygon, GraphNodes, AnyGeometry],
    img_height: int,
    img_width: int,
    with_keypoint: bool,
    num_keypoints: int,
    num_dims: int,
    coords: List[float],
) -> Union[Rectangle, Polygon, GraphNodes, None]:
    """
    Get the geometry object based on the geometry type.
    """
    if geometry_type not in GEOMETRY_CONVERTERS:
        geometry_type = detect_geometry(
            coords=coords,
            with_keypoint=with_keypoint,
            num_keypoints=num_keypoints,
            num_dims=num_dims,
        )

    if geometry_type is None:
        return None

    return GEOMETRY_CONVERTERS[geometry_type](
        img_height=img_height,
        img_width=img_width,
        coords=coords,
        num_keypoints=num_keypoints,
        num_dims=num_dims,
    )


def rectangle_to_yolo_line(
    class_idx: int,
    geometry: Rectangle,
    img_height: int,
    img_width: int,
):
    x = geometry.center.col / img_width
    y = geometry.center.row / img_height
    w = geometry.width / img_width
    h = geometry.height / img_height
    return f"{class_idx} {x:.6f} {y:.6f} {w:.6f} {h:.6f}"


def polygon_to_yolo_line(
    class_idx: int,
    geometry: Polygon,
    img_height: int,
    img_width: int,
) -> str:
    coords = []
    for point in geometry.exterior:
        x = point.col / img_width
        y = point.row / img_height
        coords.extend([x, y])
    return f"{class_idx} {' '.join(map(lambda coord: f'{coord:.6f}', coords))}"


def keypoints_to_yolo_line(
    class_idx: int,
    geometry: GraphNodes,
    img_height: int,
    img_width: int,
    max_kpts_count: int,
):
    bbox = geometry.to_bbox()
    x, y, w, h = bbox.center.col, bbox.center.row, bbox.width, bbox.height
    x, y, w, h = x / img_width, y / img_height, w / img_width, h / img_height

    line = f"{class_idx} {x:.6f} {y:.6f} {w:.6f} {h:.6f}"

    for node in geometry.nodes.values():
        node: Node
        visible = 2 if not node.disabled else 1
        line += (
            f" {node.location.col / img_width:.6f} {node.location.row / img_height:.6f} {visible}"
        )
    if len(geometry.nodes) < max_kpts_count:
        for _ in range(max_kpts_count - len(geometry.nodes)):
            line += " 0 0 0"

    return line


def convert_label_geometry_if_needed(
    label: Label,
    task_type: Literal["detection", "segmentation", "pose"],
    verbose: bool = False,
) -> List[Label]:
    if task_type == "detection":
        available_geometry_type = Rectangle
        convertable_geometry_types = [Polygon, GraphNodes, Bitmap, Polyline, AlphaMask, AnyGeometry]
    elif task_type == "segmentation":
        available_geometry_type = Polygon
        convertable_geometry_types = [Bitmap, AlphaMask, AnyGeometry]
    elif task_type == "pose":
        available_geometry_type = GraphNodes
        convertable_geometry_types = []
    else:
        raise ValueError(
            f"Unsupported task type: {task_type}. "
            "Supported types: 'detection', 'segmentation', 'pose'"
        )

    if label.obj_class.geometry_type == available_geometry_type:
        return [label]

    need_convert = label.obj_class.geometry_type in convertable_geometry_types

    if need_convert:
        new_obj_cls = label.obj_class.clone(geometry_type=available_geometry_type)
        return label.convert(new_obj_cls)

    if verbose:
        logger.warning(
            f"Label '{label.obj_class.name}' has unsupported geometry type: "
            f"{type(label.obj_class.geometry_type)}. Skipping."
        )
    return []


def label_to_yolo_lines(
    label: Label,
    img_height: int,
    img_width: int,
    class_names: List[str],
    task_type: Literal["detection", "segmentation", "pose"],
) -> List[str]:
    """
    Convert the Supervisely Label to a line in the YOLO format.
    """

    labels = convert_label_geometry_if_needed(label, task_type)
    class_idx = class_names.index(label.obj_class.name)

    lines = []
    for label in labels:
        if task_type == "detection":
            yolo_line = rectangle_to_yolo_line(
                class_idx=class_idx,
                geometry=label.geometry,
                img_height=img_height,
                img_width=img_width,
            )
        elif task_type == "segmentation":
            yolo_line = polygon_to_yolo_line(
                class_idx=class_idx,
                geometry=label.geometry,
                img_height=img_height,
                img_width=img_width,
            )
        elif task_type == "pose":
            nodes_field = label.obj_class.geometry_type.items_json_field
            max_kpts_count = len(label.obj_class.geometry_config[nodes_field])
            yolo_line = keypoints_to_yolo_line(
                class_idx=class_idx,
                geometry=label.geometry,
                img_height=img_height,
                img_width=img_width,
                max_kpts_count=max_kpts_count,
            )
        else:
            raise ValueError(f"Unsupported geometry type: {type(label.obj_class.geometry_type)}")

        if yolo_line is not None:
            lines.append(yolo_line)

    return lines


def sly_ann_to_yolo(
    ann: Annotation,
    class_names: List[str],
    task_type: Literal["detection", "segmentation", "pose"] = "detection",
) -> List[str]:
    """
    Convert the Supervisely annotation to the YOLO format.
    """

    h, w = ann.img_size
    yolo_lines = []
    for label in ann.labels:
        lines = label_to_yolo_lines(
            label=label,
            img_height=h,
            img_width=w,
            class_names=class_names,
            task_type=task_type,
        )
        yolo_lines.extend(lines)
    return yolo_lines


def sly_ds_to_yolo(
    dataset: Dataset,
    meta: ProjectMeta,
    dest_dir: Optional[str] = None,
    task_type: Literal["detection", "segmentation", "pose"] = "detection",
    log_progress: bool = False,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
) -> str:

    if progress_cb is not None:
        log_progress = False

    if log_progress:
        progress_cb = tqdm_sly(
            desc=f"Converting dataset '{dataset.short_name}' to YOLO format",
            total=len(dataset),
        ).update

    dest_dir = Path(dataset.path) / "yolo" if dest_dir is None else Path(dest_dir)
    dest_dir.mkdir(parents=True, exist_ok=True)

    # * create train and val directories
    images_dir = dest_dir / "images"
    labels_dir = dest_dir / "labels"
    train_images_dir = images_dir / "train"
    train_labels_dir = labels_dir / "train"
    val_images_dir = images_dir / "val"
    val_labels_dir = labels_dir / "val"
    for dir_path in [train_images_dir, train_labels_dir, val_images_dir, val_labels_dir]:
        dir_path.mkdir(parents=True, exist_ok=True)

    # * convert annotations and copy images
    class_names = [obj_class.name for obj_class in meta.obj_classes]
    used_names = set(os.listdir(train_images_dir)) | set(os.listdir(val_images_dir))
    for name in dataset.get_items_names():
        ann_path = dataset.get_ann_path(name)
        ann = Annotation.load_json_file(ann_path, meta)

        images_dir = val_images_dir if ann.img_tags.get("val") else train_images_dir
        labels_dir = val_labels_dir if ann.img_tags.get("val") else train_labels_dir

        img_path = Path(dataset.get_img_path(name))
        img_name = f"{dataset.short_name}_{get_file_name_with_ext(img_path)}"
        img_name = generate_free_name(used_names, img_name, with_ext=True, extend_used_names=True)
        shutil.copy2(img_path, images_dir / img_name)

        label_path = str(labels_dir / f"{img_name}.txt")
        yolo_lines = ann.to_yolo(class_names, task_type)
        if len(yolo_lines) > 0:
            with open(label_path, "w") as f:
                f.write("\n".join(yolo_lines))
        else:
            touch(label_path)

        if progress_cb is not None:
            progress_cb(1)

    # * save data config file if it does not exist
    config_path = dest_dir / "data_config.yaml"
    if not config_path.exists():
        save_yolo_config(meta, dest_dir, with_keypoint=task_type == "pose")

    return str(dest_dir)


def save_yolo_config(meta: ProjectMeta, dest_dir: str, with_keypoint: bool = False):
    dest_dir = Path(dest_dir)
    save_path = dest_dir / "data_config.yaml"
    class_names = [c.name for c in meta.obj_classes]
    class_colors = [c.color for c in meta.obj_classes]
    data_yaml = {
        "train": f"../{str(dest_dir.name)}/images/train",
        "val": f"../{str(dest_dir.name)}/images/val",
        "nc": len(class_names),
        "names": class_names,
        "colors": class_colors,
    }
    has_keypoints = any(c.geometry_type == GraphNodes for c in meta.obj_classes)
    if has_keypoints and with_keypoint:
        max_kpts_count = 0
        for obj_class in meta.obj_classes:
            if issubclass(obj_class.geometry_type, GraphNodes):
                field_name = obj_class.geometry_type.items_json_field
                max_kpts_count = max(max_kpts_count, len(obj_class.geometry_config[field_name]))
        data_yaml["kpt_shape"] = [max_kpts_count, 3]
    with open(save_path, "w") as f:
        yaml.dump(data_yaml, f, default_flow_style=None)

    logger.info(f"Data config file has been saved to {str(save_path)}")


def sly_project_to_yolo(
    project: Union[Project, str],
    dest_dir: Optional[str] = None,
    task_type: Literal["detection", "segmentation", "pose"] = "detection",
    log_progress: bool = False,
    progress_cb: Optional[Callable] = None,
):
    """
    Convert Supervisely project to YOLO format.

    :param dest_dir: Destination directory.
    :type dest_dir: :class:`str`, optional
    :param log_progress: Show uploading progress bar.
    :type log_progress: :class:`bool`
    :param progress_cb: Function for tracking conversion progress (for all items in the project).
    :type progress_cb: callable, optional
    :return: None
    :rtype: NoneType

    :Usage example:

    .. code-block:: python

        import supervisely as sly

        # Local folder with Project
        project_directory = "/home/admin/work/supervisely/source/project"

        # Convert Project to YOLO format
        sly.Project(project_directory).to_yolo(log_progress=True)
    """
    if isinstance(project, str):
        project = Project(project, mode=OpenMode.READ)

    dest_dir = Path(project.directory).parent / "yolo" if dest_dir is None else Path(dest_dir)

    dest_dir.mkdir(parents=True, exist_ok=True)
    if len(os.listdir(dest_dir)) > 0:
        raise FileExistsError(f"Directory {dest_dir} is not empty.")

    if progress_cb is not None:
        log_progress = False

    if log_progress:
        progress_cb = tqdm_sly(
            desc="Converting Supervisely project to YOLO format", total=project.total_items
        ).update

    save_yolo_config(project.meta, dest_dir, with_keypoint=task_type == "pose")

    for dataset in project.datasets:
        dataset: Dataset
        dataset.to_yolo(
            meta=project.meta,
            dest_dir=dest_dir,
            task_type=task_type,
            log_progress=log_progress,
            progress_cb=progress_cb,
        )
        logger.info(f"Dataset '{dataset.short_name}' has been converted to YOLO format.")
    logger.info(f"Project '{project.name}' has been converted to YOLO format.")



from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class TextArea(Widget):
    def __init__(
        self,
        value: str = None,
        placeholder: str = "Please input",
        rows: int = 2,
        autosize: bool = True,
        readonly: bool = False,
        widget_id=None,
    ):
        self._value = value
        self._placeholder = placeholder
        self._rows = rows
        self._autosize = autosize
        self._readonly = readonly
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "placeholder": self._placeholder,
            "rows": self._rows,
            "autosize": self._autosize,
            "readonly": self._readonly,
        }

    def get_json_state(self):
        return {"value": self._value}

    def set_value(self, value):
        self._value = value
        StateJson()[self.widget_id]["value"] = value
        StateJson().send_changes()

    def get_value(self):
        return StateJson()[self.widget_id]["value"]

    def is_readonly(self):
        return DataJson()[self.widget_id]["readonly"]

    def enable_readonly(self):
        self._readonly = True
        DataJson()[self.widget_id]["readonly"] = True
        DataJson().send_changes()

    def disable_readonly(self):
        self._readonly = False
        DataJson()[self.widget_id]["readonly"] = False
        DataJson().send_changes()



# coding: utf-8
from typing import Literal, Optional

from supervisely.api.module_api import ApiField, ModuleApiBase
from supervisely.collection.str_enum import StrEnum


class ImageAnnotationToolAction(StrEnum):
    SET_FIGURE = "figures/setFigure"
    """"""
    NEXT_IMAGE = "images/nextImage"
    """"""
    PREV_IMAGE = "images/prevImage"
    """"""
    SET_IMAGE = "images/setImage"
    """"""
    ZOOM_TO_FIGURE = "scene/zoomToObject"
    """"""
    SHOW_SUCCESS_NOTIFICATION = "app/showSuccessNotification"
    """"""
    SHOW_WARNING_NOTIFICATION = "app/showWarningNotification"
    """"""
    SHOW_ERROR_NOTIFICATION = "app/showErrorNotification"
    """"""


class ImageAnnotationToolApi(ModuleApiBase):
    def set_figure(self, session_id: str, figure_id: int):
        """Sets the figure as the current figure (selected figure) in the annotation tool.

        :param session_id: Annotation tool session id.
        :type session_id: str
        :param figure_id: Figure id.
        :type figure_id: int"""
        return self._act(
            session_id, ImageAnnotationToolAction.SET_FIGURE, {ApiField.FIGURE_ID: figure_id}
        )

    def next_image(self, session_id: str, *args, **kwargs):
        """Changes the current image in the annotation tool to the next image.

        :param session_id: Annotation tool session id.
        :type session_id: str"""
        return self._act(session_id, ImageAnnotationToolAction.NEXT_IMAGE, {})

    def prev_image(self, session_id: str, *args, **kwargs):
        """Changes the current image in the annotation tool to the previous image.

        :param session_id: Annotation tool session id.
        :type session_id: str"""
        return self._act(session_id, ImageAnnotationToolAction.PREV_IMAGE, {})

    def set_image(self, session_id: str, image_id: int):
        """Sets the image as the current image in the annotation tool.
        NOTE: The image must be in the same dataset as the current image.

        :param session_id: Annotation tool session id.
        :type session_id: str
        :param image_id: Image id in the same dataset as the current image.
        :type image_id: int"""
        return self._act(
            session_id, ImageAnnotationToolAction.SET_IMAGE, {ApiField.IMAGE_ID: image_id}
        )

    def zoom_to_figure(self, session_id: str, figure_id: int, zoom_factor: Optional[float] = 1):
        """Zooms the scene to the figure with the given id and zoom factor.

        :param session_id: Annotation tool session id.
        :type session_id: str
        :param figure_id: Figure id.
        :type figure_id: int
        :param zoom_factor: Zoom factor. Default is 1.
        :type zoom_factor: float, optional"""
        return self._act(
            session_id,
            ImageAnnotationToolAction.ZOOM_TO_FIGURE,
            {ApiField.FIGURE_ID: figure_id, ApiField.ZOOM_FACTOR: zoom_factor},
        )

    def show_notification(
        self,
        session_id: str,
        message: str,
        notification_type: Literal["success", "warning", "error"],
        duration: Optional[int] = None,
    ):
        """Shows a notification in the annotation tool.

        :param session_id: Annotation tool session id.
        :type session_id: str
        :param message: Notification message.
        :type message: str
        :param notification_type: Notification type. One of "success", "warning", "error".
        :type notification_type: Literal["success", "warning", "error"]
        :param duration: Notification duration in milliseconds. Default is None.
        :type duration: int, optional
        :raises ValueError: If notification_type is invalid.
        """
        actions = {
            "success": ImageAnnotationToolAction.SHOW_SUCCESS_NOTIFICATION,
            "warning": ImageAnnotationToolAction.SHOW_WARNING_NOTIFICATION,
            "error": ImageAnnotationToolAction.SHOW_ERROR_NOTIFICATION,
        }
        try:
            action = actions[notification_type]
        except KeyError:
            raise ValueError(
                f"Invalid notification type, expected one of {list(actions.keys())}, "
                f"got {notification_type}"
            )

        payload = {ApiField.MESSAGE: message, ApiField.DURATION: duration}
        return self._act(session_id, action, payload)

    def _act(self, session_id: str, action: ImageAnnotationToolAction, payload: dict):
        """ """
        data = {
            ApiField.SESSION_ID: session_id,
            ApiField.ACTION: str(action),
            ApiField.PAYLOAD: payload,
        }
        resp = self._api.post("/annotation-tool.run-action", data)
        return resp.json()

    # {
    #     "sessionId": "940c4ec7-3818-420b-9277-ab3c820babe5",
    #     "action": "scene/setViewport",
    #     "payload": {
    #         "viewport": {
    #             "offsetX": -461, # width
    #             "offsetY": -1228, # height
    #             "zoom": 1.7424000000000024
    #         }
    #     }
    # }

    # {
    #     "sessionId": "940c4ec7-3818-420b-9277-ab3c820babe5",
    #     "action": "scene/zoomToObject",
    #     "payload": {
    #         "figureId": 22129,
    #         "zoomFactor": 1.5
    #     }
    # }



# isort: skip_file

# Project
from supervisely.convert.image.coco.coco_helper import sly_project_to_coco as project_to_coco
from supervisely.convert.image.yolo.yolo_helper import sly_project_to_yolo as project_to_yolo
from supervisely.convert.image.pascal_voc.pascal_voc_helper import sly_project_to_pascal_voc as project_to_pascal_voc

# Dataset
from supervisely.convert.image.coco.coco_helper import sly_ds_to_coco as dataset_to_coco
from supervisely.convert.image.yolo.yolo_helper import sly_ds_to_yolo as dataset_to_yolo
from supervisely.convert.image.pascal_voc.pascal_voc_helper import sly_ds_to_pascal_voc as dataset_to_pascal_voc

# Image Annotations
from supervisely.convert.image.coco.coco_helper import sly_ann_to_coco as annotation_to_coco
from supervisely.convert.image.yolo.yolo_helper import sly_ann_to_yolo as annotation_to_yolo
from supervisely.convert.image.pascal_voc.pascal_voc_helper import sly_ann_to_pascal_voc as annotation_to_pascal_voc



from types import SimpleNamespace

docs_url = (
    "https://docs.supervisely.com/neural-networks/model-evaluation-benchmark/semantic-segmentation"
)

clickable_label = """
> <span style="color: #5a6772">
>     Click on the chart to explore corresponding images.
> </span>
"""

markdown_header = """
<h1>{}</h1>

<div class="model-info-block">
    <div>Created by <b>{}</b></div>
    <div><i class="zmdi zmdi-calendar-alt"></i><span>{}</span></div>
</div>
"""


markdown_overview = """
- **Model**: {}
- **Checkpoint**: {}
- **Architecture**: {}
- **Task type**: {}
- **Runtime**: {}
- **Checkpoint file**: <a class="checkpoint-url" href="{}" target="_blank">{}</a>
- **Ground Truth project**: <a href="/projects/{}/datasets" target="_blank">{}</a>, {}{}
{}

Learn more about Model Benchmark, implementation details, and how to use the charts in our <a href="{}" target="_blank">Technical Report</a>.
"""


markdown_key_metrics = """## Key Metrics

We provide a comprehensive model performance analysis using a set of metrics, including both basic (precision, recall, f1-score, IoU, etc.) and advanced (boundary IoU, error over unio decomposition, etc.) metrics.

- **Pixel accuracy**: reflects the percent of image pixels which were correctly classified.
- **Precision**: reflects the number of correctly predicted positive segmentations divided by the total number of predicted positive segmentations.
- **Recall**: reflects the number of correctly predicted positive segmentations divided by the number of all samples that should have been segmented as positive.
- **F1-score**: reflects the tradeoff between precision and recall. It is equivalent to the Dice coefficient and calculated as a harmonic mean of precision and recall.
- **Intersection over union (IoU, also known as the Jaccard index)**: measures the overlap between ground truth mask and predicted mask. It is calculated as the ratio of the intersection of the two masks areas to their combined areas.
- **Boundary intersection over union**: a segmentation consistency measure that first computes the sets of ground truth  and predicted masks pixels that are located within the distance d from each contour and then computes intersection over union of these two sets. Pixel distance parameter d (pixel width of the boundary region) controls the sensitivity of the metric, it is usually set as 2% of the image diagonal for normal resolution images and 0.5% of the image diagonal for high resolution images.
- **Error over union and its components (boundary, extent, segment)**: a metric opposite to intersection over union and can be interpreted as what the model lacked in order to show the perfect performance with IoU = 1. It reflects the ratio of incorrectly segmented pixels of ground truth and predicted masks to their combined areas. It is usually decomposed into boundary, extent and segment errors over union in order to get exhaustive information about the model's strengths and weaknesses.
- **Renormalized error over union**: postprocessed variant of error over union which takes into consideration cause and effect relationships between different types of segmentation errors.
"""

markdown_explorer = """## Explore Predictions
This section contains visual comparison of model predictions and ground truth masks. Sometimes a simple visualization can be more informative than any performance metric.

> Click on the image to view the **Original Image** with **Ground Truth** and **Prediction** annotations side-by-side. 
"""


markdown_predictions_table = """### Prediction details for every image

Table of per image metrics allows to get performance metrics for every image. It can be helpful when there is a need to find the most problematic images where the model performed worst.

**Example**: you can sort by **Pixel accuracy** in ascending order to find images where the model performed worst in terms of pixel-wise accuracy.

> Click on the row to view the **Original Image** with **Ground Truth** and **Prediction** annotations side-by-side. 
"""

markdown_iou = """## Intersection & Error Over Union

The pie chart below demonstrates what the model lacked in order to show the perfect performance with IoU = 1. It is done with the help of Error over Union (EoU) decomposition into boundary, extent and segment errors over union. This chart helps to draw conclusions on the model's strongest and weakest sides.
"""


markdown_renormalized_error_ou = """## Renormalized Error Over Union

The chart below is dedicated to decomposition of postprocessed variant of error over union which takes into consideration cause and effect relationships between different types of segmentation errors. Error over union decomposition has its own pitfalls. It is important to understand that models which tend to produce segment errors (when entire segments are mispredicted and there is no intersection between ground truth and predicted mask) will face less occasions to produce boundary and extent errors - as a result, boundary and extent error over union values will be underestimated.

In terms of localization, segment error is more fundamental than extent, while extent error is more fundamental than boundary. In order to overcome this problem, renormalized error over union proposes a slightly different calculation method - by removing more fundamental errors from the denominator - read more in our <a href="{}" target="_blank">technical report</a>
""".format(
    docs_url
)


markdown_eou_per_class = """## Classwise Segmentation Error Analysis

This section contains information about classwise segmentation error decomposition. Each column of the chart represents a certain class from the training dataset, demonstrating model performance in terms of segmenting this specific class on images and what model lacked in order to show the perfect performance. All classes are sorted in IoU descending order.
"""

markdown_confusion_matrix = """## Confusion Matrix

The confusion matrix below reveals which classes the model commonly confuses with each other.

- **Each row** of the matrix corresponds to the actual instances of a class.
- **Each column** corresponds to the instances as predicted by the model.
- **The diagonal elements** of the matrix represent correctly predicted instances.
- By examining the **off-diagonal elements**, you can see if the model is confusing two classes by frequently mislabeling one as the other.

"""

markdown_frequently_confused = """## Frequently Confused Classes

The bar chart below reveals pairs of classes which were most frequently confused by the model. Each column of the chart demonstrates the probability of confusion of a given pair of classes. It is necessary to remember that this probability is not symmetric: the probability of confusing class A with class B is not equal to the probability of confusing class B with class A.
"""

markdown_frequently_confused_empty = """### Frequently Confused Classes

No frequently confused class pairs found
"""

markdown_speedtest_intro = """## Inference Speed

This is a speed test benchmark for this model. The model was tested with the following configuration:

- **Device**: {}
- **Hardware**: {}
- **Runtime**: {}

The table below shows the speed test results. For each test, the time taken to process one batch of images is shown, as well as the model's throughput (i.e, the number of images processed per second, or FPS). Results are averaged across **{}** iterations.
"""

markdown_batch_inference = """
This chart shows how the model's speed changes with different batch sizes. As the batch size increases, you can observe an increase in FPS (images per second).
"""

markdown_acknowledgement = """---
### Acknowledgement

[1] Maximilian Bernhard, Roberto Amoroso, Yannic Kindermann, Lorenzo Baraldi, Rita Cucchiara, Volker Tresp, Matthias Schubert. <a href="https://openaccess.thecvf.com/content/WACV2024/html/Bernhard_Whats_Outside_the_Intersection_Fine-Grained_Error_Analysis_for_Semantic_Segmentation_WACV_2024_paper.html" target="_blank">What's Outside the Intersection? Fine-grained Error Analysis for Semantic Segmentation Beyond IoU.</a> In Proceedings of the IEEE / CVF Conference on Computer Vision and Pattern Recognition, pages 969 - 977, 2024.

[2] Bowen Cheng, Ross Girshick, Piotr Dollar, Alexander C. Berg, Alexander Kirillov. <a href="https://arxiv.org/abs/2103.16562" target="_blank">Boundary IoU: Improving object-centric image segmentation evaluation.</a> In Proceedings of the IEEE / CVF Conference on Computer Vision and Pattern Recognition, pages 15334 - 15342, 2021.
"""
